#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-12-28
# File  : fritzbox_smarthome_power_meter.py (check plugin)
#
#
# import time
from time import localtime, time as time_now
from typing import Dict
from os import environ
from pathlib import Path
from json import loads, dumps

from cmk.checkers import plugin_contexts
from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    Metric,
    Result,
    Service,
    State,
    check_levels,
    register,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import CheckResult, DiscoveryResult
from cmk.utils.render import physical_precision
from cmk.base.plugins.agent_based.utils.fritzbox_smarthome import AvmSmartHomeDevice

OMD_ROOT = environ["OMD_ROOT"]


class FritzBoxValueStore:
    """
    Persistent kind of value_store. Located under ~/var/fritzbox_smarthome/{hostname}_{service_description}
    """
    def __init__(self):
        self._host_name = plugin_contexts.host_name()
        self._service_description = plugin_contexts.service_description().replace(' ', '_')
        self._file = Path(f'{OMD_ROOT}/var/firtzbox_smarthome/{self._host_name}_{self._service_description}')
        self._counters = {}
        if self._file.exists():
            self._counters = loads(self._file.read_text())
        else:
            self._counters = {}

    def get(self, key: str, default_value: int | str | float | bool | None = None):
        """
        Get a value from the value store
        Args:
            key: the key in the value store
            default_value: the default value to return if key not found

        Returns:
            The value for key from the value store. If key not exists the default_value or None
        """
        return self._counters.get(key, default_value)

    def set(self, key: str, value: str | int | float | None) -> None:
        """
        Set a value in the value stare. Use "None" as value to remove a key.
        Args:
            key: the key in the value store
            value: the value to set
        """
        if value is None:
            try:
                self._counters.pop(key)
            except KeyError:
                pass
        else:
            self._counters[key] = value

    def save(self) -> None:
        """
        Save the value_store as json to file
        """
        self._file.parent.mkdir(exist_ok=True, parents=True)
        self._file.write_text(dumps(self._counters))


def discovery_fritzbox_smarthome_voltage_single(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if isinstance(section, AvmSmartHomeDevice):
        if section.power_meter and section.power_meter.voltage is not None:
            yield Service()


def discovery_fritzbox_smarthome_voltage_multiple(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if not isinstance(section, AvmSmartHomeDevice):
        for device_id, device in section.items():
            if device.power_meter and device.power_meter.voltage is not None:
                yield Service(item=str(device_id))


def check_fritzbox_smarthome_voltage_single(
        params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if not isinstance(section, AvmSmartHomeDevice):
        return
    if section.power_meter and section.power_meter.voltage is not None:
        yield from check_levels(
            label='Voltage',
            levels_lower=params.get('levels_lower'),
            levels_upper=params.get('levels'),
            metric_name='voltage',
            render_func=lambda x: f'{x}V',
            value=section.power_meter.voltage,
        )


def check_fritzbox_smarthome_voltage_multiple(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if isinstance(section, Dict):
        try:
            yield from check_fritzbox_smarthome_voltage_single(params, section[item])
        except KeyError:
            return


register.check_plugin(
    name='fritzbox_smarthome_voltage_single',
    service_name='Voltage',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_voltage_single,
    check_function=check_fritzbox_smarthome_voltage_single,
    check_ruleset_name='voltage_single',
    check_default_parameters={}
)

register.check_plugin(
    name='fritzbox_smarthome_voltage_multiple',
    service_name='Smarthome Voltage %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_voltage_multiple,
    check_function=check_fritzbox_smarthome_voltage_multiple,
    check_ruleset_name='voltage',
    check_default_parameters={}
)


def discovery_fritzbox_smarthome_power_single(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if isinstance(section, AvmSmartHomeDevice):
        if section.power_meter and section.power_meter.power is not None:
            yield Service()


def discovery_fritzbox_smarthome_power_multiple(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if not isinstance(section, AvmSmartHomeDevice):
        for device_id, device in section.items():
            if device.power_meter and device.power_meter.power is not None:
                yield Service(item=str(device_id))


def check_fritzbox_smarthome_power_single(
        params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if not isinstance(section, AvmSmartHomeDevice):
        return

    if section.power_meter and section.power_meter.power is not None:
        yield from check_levels(
            value=section.power_meter.power,
            metric_name='power',
            label='Power',
            render_func=lambda x: f'{x}W',
            levels_upper=params.get('levels_upper'),
            levels_lower=params.get('levels_lower'),
        )


def check_fritzbox_smarthome_power_multiple(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if isinstance(section, Dict):
        try:
            yield from check_fritzbox_smarthome_power_single(params, section[item])
        except KeyError:
            return


register.check_plugin(
    name='fritzbox_smarthome_power_single',
    service_name='Power',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_power_single,
    check_function=check_fritzbox_smarthome_power_single,
    check_ruleset_name='epower_single',
    check_default_parameters={}
)

register.check_plugin(
    name='fritzbox_smarthome_power_multiple',
    service_name='Smarthome Power %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_power_multiple,
    check_function=check_fritzbox_smarthome_power_multiple,
    check_ruleset_name='epower',
    check_default_parameters={}
)


def discovery_fritzbox_smarthome_energy_single(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if isinstance(section, AvmSmartHomeDevice):
        if section.power_meter and section.power_meter.energy is not None:
            yield Service()


def discovery_fritzbox_smarthome_energy_multiple(
        section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> DiscoveryResult:
    if not isinstance(section, AvmSmartHomeDevice):
        for device_id, device in section.items():
            if device.power_meter and device.power_meter.energy is not None:
                yield Service(item=str(device_id))


def _cost_period_x(
        value_store: FritzBoxValueStore,
        period_name: str,
        current_period: int,
        rate_name: str,
        metric_name: str,
        last_reading: float,
        precision: int,
        message: str,
        cost_kwh: float,
) -> CheckResult:
    # reset all
    # value_store[rate_name] = 0

    if stored_period := value_store.get(key=period_name):
        value_store.set(key=period_name, value=current_period)

    cost = last_reading / 1000 * cost_kwh

    if current_period != stored_period:
        value_store.set(key=period_name, value=current_period)
        value_store.set(key=rate_name, value=cost)
    else:
        cost = value_store.get(key=rate_name, default_value=0) + cost
        value_store.set(key=rate_name, value=cost)

    cost = round(cost, precision)

    yield Result(
        state=State.OK,
        notice=message.replace('__value__', f'{cost:.4f}')
    )
    yield Metric(name=metric_name, value=cost)


def check_fritzbox_smarthome_energy_single(
        params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if not isinstance(section, AvmSmartHomeDevice):
        return

    energy = None
    time_span = None
    value_store = FritzBoxValueStore()

    if section.power_meter and section.power_meter.power is not None:
        if not (last_timestamp := value_store.get('last_timestamp')):
            value_store.set(key='last_timestamp', value=time_now())
        else:
            time_span = time_now() - last_timestamp
            value_store.set(key='last_timestamp', value=time_now())

            energy = section.power_meter.power / 3600 * time_span

            yield from check_levels(
                value=energy,
                metric_name='energy_current',
                label='Consumption since last reading',
                render_func=lambda x: f'{physical_precision(v=x, precision=3, unit_symbol="Wh")} (estimated)',
                levels_lower=params.get('levels_lower'),
                levels_upper=params.get('levels_upper'),
            )
            yield Metric(name='energy_timespan', value=time_span)

    if section.power_meter and section.power_meter.energy is not None:
        yield Result(
            state=State.OK,
            summary=f'Consumption total'
                    f': {physical_precision(v=section.power_meter.energy, precision=3, unit_symbol="Wh")}'
        )
        yield Metric(name='energy_total', value=section.power_meter.energy)

        if (cost_kwh := params.get('cost_kwh', None)) is not None:
            _currency = {
                'CHF': 'fr',  # swiss franc
                'CZK': 'Kč',  # Czech koruna
                'DKK': 'kr.',  # Danish krone
                'EUR': '€',
                'GBP': '£',
                'JPY': '¥',
                'PLN': 'zł',  # Polish złoty
                'USD': '$',
            }
            cost_kwh, unit_sign = cost_kwh
            unit_sign = _currency.get(unit_sign, "€")

            cost = section.power_meter.energy / 1000 * cost_kwh

            yield Result(
                state=State.OK,
                summary=f'Cost total: {cost:.2f}{unit_sign} ({cost_kwh:.2f}{unit_sign}/kWh)',
            )

            if energy is not None and time_span is not None:
                yield Result(state=State.OK, notice=' ')
                yield Result(state=State.OK, notice='Cost for this:')
                loca_time = localtime()
                yield from _cost_period_x(
                    value_store=value_store,
                    period_name='current_hour',
                    current_period=loca_time.tm_hour,
                    rate_name='cost_this_hour',  # don't reuse -> cost_per_hour
                    metric_name='cost_per_hour',
                    last_reading=energy,
                    precision=6,
                    message=f'Hour_: __value__{unit_sign}',
                    cost_kwh=cost_kwh,
                )
                yield from _cost_period_x(
                    value_store=value_store,
                    period_name='current_day',
                    current_period=loca_time.tm_mday,
                    rate_name='cost_this_day',  # don't reuse -> cost_per_day
                    metric_name='cost_per_day',
                    last_reading=energy,
                    precision=4,
                    message=f'Day__: __value__{unit_sign}',
                    cost_kwh=cost_kwh,
                )
                yield from _cost_period_x(
                    value_store=value_store,
                    period_name='current_month',
                    current_period=loca_time.tm_mon,
                    rate_name='cost_this_month',  # don't reuse -> cost_per_month
                    metric_name='cost_per_month',
                    last_reading=energy,
                    precision=4,
                    message=f'Month: __value__{unit_sign}',
                    cost_kwh=cost_kwh,
                )
                yield from _cost_period_x(
                    value_store=value_store,
                    period_name='current_year',
                    current_period=loca_time.tm_year,
                    rate_name='cost_this_year',  # don't reuse -> cost_per_year
                    metric_name='cost_per_year',
                    last_reading=energy,
                    precision=4,
                    message=f'Year_: __value__{unit_sign}',
                    cost_kwh=cost_kwh,
                )
                yield Result(state=State.OK, notice=' ')

    value_store.save()

    if energy:
        yield Result(
            state=State.OK,
            notice=f'All estimates based on the assumption, of a constant '
                   f'power usage of {section.power_meter.power}W '
                   f'for the last {time_span:.2f} seconds'
        )


def check_fritzbox_smarthome_energy_multiple(
        item, params, section: AvmSmartHomeDevice | Dict[str, AvmSmartHomeDevice]
) -> CheckResult:
    if isinstance(section, Dict):
        try:
            yield from check_fritzbox_smarthome_energy_single(params, section[item])
        except KeyError:
            return


register.check_plugin(
    name='fritzbox_smarthome_energy_single',
    service_name='Energy',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_energy_single,
    check_function=check_fritzbox_smarthome_energy_single,
    check_ruleset_name='energy_single',
    check_default_parameters={}
)

register.check_plugin(
    name='fritzbox_smarthome_energy_multiple',
    service_name='Smarthome Energy %s',
    sections=['fritzbox_smarthome'],
    discovery_function=discovery_fritzbox_smarthome_energy_multiple,
    check_function=check_fritzbox_smarthome_energy_multiple,
    check_ruleset_name='energy_multiple',
    check_default_parameters={}
)
